import numpy as np

from cython cimport boundscheck, wraparound
from cython.view cimport contiguous
from cython.parallel import prange
cimport numpy as cnp


# Separate fused types for input and output to allow for the
# combination to occur.
ctypedef fused inp_t:
    int
    unsigned int
    float
    double

ctypedef fused outp_t:
    int
    unsigned int
    float
    double


@wraparound(False)
@boundscheck(False)
def transpose_constant(outp_t[:, :, :, ::contiguous] dst,
                       inp_t[:, :, :, ::contiguous] src,
                       int num_threads=0):
    """
    Optimized and parallelized constant transposition.

    Will work for any 4D array, but the iteration order is handpicked
    to be fast for AGIPD constants.
    """

    if not (dst.shape[0] == src.shape[3] and dst.shape[1] == src.shape[2] and
            dst.shape[2] == src.shape[1] and dst.shape[3] == src.shape[0]):
        raise ValueError(f'dst shape {np.array(dst.shape)[:4]} must be '
                         f'reversed src shape {np.array(src.shape)[:4]}')

    cdef int y, x, c, g

    # Optimized iteration (on INTEL) order via profiling, do not change
    # and don't ask me why!
    for c in prange(src.shape[2], nogil=True, num_threads=num_threads):
        for y in range(src.shape[0]):
            for x in range(src.shape[1]):
                for g in range(src.shape[3]):
                    dst[g, c, x, y] = <outp_t>src[y, x, c, g]


@boundscheck(False)
@wraparound(False)
def histogram(cnp.ndarray[cnp.float32_t, ndim=2] data, range=(0,1), int bins=20,
           cnp.ndarray[cnp.float32_t, ndim=2] weights=None):
    """
    Calculate N histograms along axis 0 with equidistant binning.
    N is a data size along axis 1.
    Return histograms and bin edges.
    """

    cdef cnp.ndarray[cnp.float32_t, ndim=2] ret
    cdef double min, max
    min = range[0]
    max = range[1]

    ret = np.zeros((bins,data.shape[1]), dtype=np.float32)
    cdef double bin_width = (max - min) / bins
    cdef double x

    if weights is not None:
        assert (<object>weights).shape == (<object>data).shape,\
            "data & weights must have matching shape"

    for j in xrange(data.shape[1]):
        for i in xrange(data.shape[0]):
            x = (data[i,j] - min) / bin_width
            if 0.0 <= x < bins:
                if weights is None:
                    ret[<int>x,j] += 1.0
                else:
                    ret[<int>x,j] += weights[i,j]
    return ret, np.linspace(min, max, bins+1)


@boundscheck(False)
@wraparound(False)
def histogram2d(cnp.ndarray[float, ndim=1] data_x, cnp.ndarray[float, ndim=1] data_y,
                range=((0,1), (0,1)), bins=(20, 20)):
    """
    Calculate 2D histogram with equidistant binning.
    Return histogram and bin edges.
    """

    cdef cnp.ndarray[cnp.int32_t, ndim=2] ret
    cdef double min_x, max_x, min_y, max_y
    cdef int bins_x, bins_y
    min_x = range[0][0]
    max_x = range[0][1]
    min_y = range[1][0]
    max_y = range[1][1]
    bins_x = bins[0]
    bins_y = bins[1]

    ret = np.zeros((bins_x, bins_y), dtype=np.int32)

    cdef double bin_width_x = (max_x - min_x) / bins_x
    cdef double bin_width_y = (max_y - min_y) / bins_y
    cdef double x, y

    assert (<object>data_x).shape == (<object>data_y).shape,\
        "data_x & data_y must have matching shape"

    for i in xrange(data_x.shape[0]):
        x = (data_x[i] - min_x) / bin_width_x
        y = (data_y[i] - min_y) / bin_width_y
        if x >= 0.0 and x < bins_x and y>=0.0 and y<bins_y:
            ret[<int>x, <int>y] += 1

    return ret, np.linspace(min_x, max_x, bins_x+1), np.linspace(min_y, max_y, bins_y+1)


ctypedef fused choices_t:
    cnp.float32_t
    cnp.int32_t

@boundscheck(False)
@wraparound(False)
def gain_choose(cnp.ndarray[cnp.uint8_t, ndim=3] a, cnp.ndarray[choices_t, ndim=4] choices):
    """Specialised fast equivalent of np.choose(), to select data for a per-pixel gain stage"""
    cdef int i, j, k
    cdef cnp.uint8_t v
    cdef cnp.ndarray[choices_t, ndim=3] out
    out = np.zeros_like(a, dtype=(<object>choices).dtype)

    assert (<object>choices).shape == (3,) + (<object>a).shape

    with nogil:
        for i in range(a.shape[0]):
            for j in range(a.shape[1]):
                for k in range(a.shape[2]):
                    v = a[i, j, k]
                    out[i, j, k] = choices[v, i, j, k]

    return out


@boundscheck(False)
@wraparound(False)
def cm_correction(float[:, :, :] arr, unsigned short[:] cellid,
                  float lower, float upper, float fraction):
    """Apply one iteration of common-mode correction

    The common-mode correction shifts the position of the noise peak for
    a slice of pixels to zero. The position is estimated as mean signal of
    pixels in a range between `lower` and `upper` boundaries. If the noise
    peak located in this boundaries then the mean value works as a robust
    estimator, which converges to the true noise peak position with
    iterations. The correction is applied only if the number of pixels in
    the given range above the `fraction` of the total number of the pixels
    in the slice.

    This function performs consequently cells common-mode correction and
    ASIC common-mode correction.
    Cells correction is calculated across trains and groups of 32 cells.
    ASIC correction is calculated across the pixes in ASIC on a single image.

    To converge this function should be iterated.
    
    This function performs correction in-place, altering `arr`.

    :param arr: array of images cropped to a single ASIC and
        stacked together over trains and pulses in the first dimension
    :param cellid: array of cell IDs, must have the length equal to
        the number of images
    :param lower: the lower signal value in ADU to consider pixel as a dark
    :param upper: the upper signal value in ADU to consider pixel as a dark
    :param fraction: the fraction of the dark pixels in the slice which is
        considered to be enough to apply the common-mode correction to this
        slice
    """
    cdef long long nfrm = arr.shape[0]
    cdef long long nx = arr.shape[1]
    cdef long long ny = arr.shape[2]

    cdef float asic_thr = fraction * nx * ny

    cdef double[:, :, :] crng_sum = np.zeros([11, nx, ny], dtype=np.float64)
    cdef long long[:, :, :] crng_cnt = np.zeros([11, nx, ny], dtype=np.int64)
    cdef long long[:, :, :] crng_tot = np.zeros([11, nx, ny], dtype=np.int64)

    cdef long long asic_cnt, cnt, tot, i, l, m
    cdef int row
    cdef double asic_sum
    cdef float value
    cdef bint z
    cdef bint used_row[11]

    with nogil:
        for row in range(11):
            used_row[row] = False
        # sum and count intensities over cell rows and trains
        # the result has dimensionality [11, 64, 64]
        for i in range(nfrm):
            row = cellid[i] // 32
            used_row[row] = True
            for l in range(nx):
                for m in range(ny):
                    value = arr[i, l, m]
                    z = (lower <= value) and (value <= upper)
                    crng_sum[row, l, m] += z * value
                    crng_cnt[row, l, m] += z
                    crng_tot[row, l, m] += 1

        # find average values if there are more values in the interval
        # than `fraction`
        for row in range(11):
            if used_row[row]:
                for l in range(nx):
                    for m in range(ny):
                        tot = crng_tot[row, l, m]
                        cnt = crng_cnt[row, l, m]
                        z = cnt < (fraction * tot)
                        crng_sum[row, l, m] = (
                            .0 if z else (crng_sum[row, l, m] / cnt))

        # subtract mean value from the intensities
        for i in range(nfrm):
            row = cellid[i] // 32
            for l in range(nx):
                for m in range(ny):
                    arr[i, l, m] -= crng_sum[row, l, m]

        # over ASIC pixels
        for i in range(nfrm):
            asic_sum = .0
            asic_cnt = 0
            # sum and count
            for l in range(nx):
                for m in range(ny):
                    value = arr[i, l, m]
                    z = (lower <= value) and (value <= upper)
                    asic_sum += z * value
                    asic_cnt += z

            # find average values if there are more values in the interval
            # than `fraction`
            z = asic_cnt < asic_thr
            asic_sum = .0 if z else (asic_sum / asic_cnt)

            # subtract mean value from the intensities
            for l in range(nx):
                for m in range(ny):
                    arr[i, l, m] -= asic_sum
